unit vvvImageGr;

{$mode objfpc}{$H+}

interface

uses
 Classes, SysUtils, Graphics, LCLType, LCLIntf, vvvOCR;

type
 TbmpGrayImage = class(TgrayImage)
 public
  constructor Create(src: TGraphic); overload;
  procedure CopyGraphic(src: TGraphic);
  procedure CopyGreen(src: TGraphic);
  procedure DrawToCanvas(canvas: TCanvas);
  function asBitmap: TBitmap;
 end;

implementation

constructor TbmpGrayImage.Create(src: TGraphic);
begin
 inherited Create();
 CopyGraphic(src);
end;

procedure TbmpGrayImage.CopyGraphic(src: TGraphic);
var
 bmp: TBitmap;
 x, y, c, pw{$ifdef windows}, i{$endif}: integer;
 d: PByte;
begin
 bmp:= TBitmap.Create;
 bmp.PixelFormat:= pf24bit;
 bmp.Width:= src.Width;
 bmp.Height:= src.Height;
 bmp.Canvas.Draw(0, 0, src);
 Width:= bmp.Width;
 Height:= bmp.Height;
 d:= bmp.RawImage.Data;
 pw:= bmp.RawImage.Description.BytesPerLine div bmp.Width;
  {$ifdef windows}
  for y:= 0 to Height - 1 do begin
   d:= bmp.ScanLine[y];
   for x:= 0 to Width - 1 do begin
    c:= 0;
    for i:= 0 to pw - 1 do inc(c, d[x * pw + i]);
    c:= c div 3;
    Pixel[x, y]:= c and $FF;
   end;
  end;
 {$else}
 for x:= 0 to Width * Height  - 1 do begin
  c:= 0;
  for y:= 0 to 2 do inc(c, d[x * pw + y]);
  c:= c div 3;
  Data[x]:= c and $FF;
 end;
 {$endif}
 bmp.FreeImage;
 bmp.Free;
 DataChanged;
end;

procedure TbmpGrayImage.CopyGreen(src: TGraphic);
var
 bmp: TBitmap;
 x, pw: integer;
 d: PByte;
begin
 bmp:= TBitmap.Create;
 bmp.PixelFormat:= pf24bit;
 bmp.Width:= src.Width;
 bmp.Height:= src.Height;
 bmp.Canvas.Draw(0, 0, src);
 Width:= bmp.Width;
 Height:= bmp.Height;
 d:= bmp.RawImage.Data;
 pw:= bmp.RawImage.Description.BytesPerLine div bmp.Width;
 for x:= 0 to Width * Height  - 1 do Data[x]:= d[x * pw + 1];
 bmp.FreeImage;
 bmp.Free;
 DataChanged;
end;

procedure TbmpGrayImage.DrawToCanvas(canvas: TCanvas);
var
 bmp: TBitmap;
begin
 bmp:= asBitmap;
 try
  canvas.Draw(0, 0, bmp);
 finally
  bmp.Free;
 end;
end;

function TbmpGrayImage.asBitmap: TBitmap;
const
 mul: array [boolean] of byte = (1, 255);
var
 x, y, pw{$ifdef windows}, i{$endif}: integer;
 d: PByte;
begin
 Result:= TBitmap.Create;
 Result.PixelFormat:= pf24bit;
 Result.Width:= Width;
 Result.Height:= Height;
 Result.BeginUpdate;
 d:= Result.RawImage.Data;
 pw:= Result.RawImage.Description.BytesPerLine div Result.Width;
 {$ifdef windows}
  for y:= 0 to Height - 1 do begin
   d:= Result.ScanLine[y];
   for x:= 0 to Width - 1 do for i:= 0 to pw - 1 do d[x * pw + i]:= Data[y * Width + x] * mul[OneBit];
  end;
 {$else}
 for x:= 0 to Height * Width - 1 do begin
  for y:= 0 to 2 do d[x * pw + y]:= Data[x] * mul[OneBit];
 end;
 {$endif}
 Result.EndUpdate;
end;

end.

