unit vvvPassport;

{$mode objfpc}{$H+}

interface

uses
 Classes, SysUtils, vvvOCR;

type
 TpassportFinder = class(TbasicOCR)
 private
  fSHeightN, fSHeightX: byte;
  fSCntr: word;
  fFound: boolean;
  fObjectsFound: boolean;
  fDocType: ansiString;
  fIssued: ansiString;
  fSurname: ansiString;
  fNumber: ansiString;
  fNationality: ansiString;
  fBirthday: TDateTime;
  fSex: ansiChar;
  fExpiration: TDateTime;
  fPersonalNumber: ansiString;
  procedure CheckSymbols(const Sender: TgrayImage; const r: TRect; var stop: boolean);
 public
  SPositions: array [word] of TRect;
  badVideo: boolean;
  property SCntr: word read fSCntr;
  property Found: boolean read fFound;
  property ObjectsFound: boolean read fObjectsFound;
  property DocType: ansiString read fDocType;
  property Issued: ansiString read fIssued;
  property Surname: ansiString read fSurname;
  property Number: ansiString read fNumber;
  property Nationality: ansiString read fNationality;
  property Birthday: TDateTime read fBirthday;
  property Sex: ansiChar read fSex;
  property Expiration: TDateTime read fExpiration;
  property PersonalNumber: ansiString read fPersonalNumber;
 public
  constructor Create;
  function Search(const img: TgrayImage; const r: TRect): boolean; override;
 end;

implementation

constructor TpassportFinder.Create;
begin
 inherited;
 badVideo:= False;
 minWeight:= 50;
end;

function TpassportFinder.Search(const img: TgrayImage; const r: TRect): boolean;
const
 fl: array [boolean, 0..1] of boolean = ((False, True), (True, False));
 fli: array [boolean] of byte = (1, 0);
type
 tLine = record
  count: byte;
  minHeight,
  avgHeight: word;
  pos: array [byte] of TRect;
 end;
var
 lines: array of tLine;
 fc, c: integer;
 df: TFormatSettings;

 function rectsK(r1, r2: TRect): extended;
 begin
  Result:=
   (r1.Top + r1.Height / 2 - r2.Top - r2.Height / 2) /
   (r2.Left + r2.Width / 2 - r1.Left - r1.Width / 2)
 end;

 function checksumMod(var s: ansiString): byte;
 const
  mul: array [0..2] of byte = (7,3,1);
 var
  counter: integer;
  sm: integer;
 begin
  sm:= 0;
  for counter:= 1 to length(s) do
   if s[counter] in ['A'..'Z'] then inc(sm,(10 + ord(s[counter]) - 65) * mul[(counter - 1) mod 3])
   else if s[counter] in ['1'..'9'] then inc(sm, (ord(s[counter]) - 48) * mul[(counter - 1) mod 3]);
  Result:= (sm mod 10)
 end;

 function checksumSIsOk(x: byte; checksum: ansiChar): boolean; overload;
 begin
  if checksum = 'O' then checksum:= '0';
  if checksum = '<' then checksum:= '0';
  Result:= x mod 10 = ord(checksum) - 48;
 end;

 function checksumIsOk(var s: ansiString; checksum: ansiChar): boolean; overload;
 begin
  Result:= checksumSIsOk(checksumMod(s), checksum);
 end;

 function tryRecognize: boolean;
 var
  lc, c, n: integer;
  dx, dy, dh, ah: word;
  rLines: array [0..1] of tLine;
  rTexts: array [0..1] of ansiString;
  s, t: ansiString;
  f: boolean;
  p: TRect;
  rc: char;
  img1: TgrayImage;
  angle: integer;

 procedure check(r1, r2: TRect);
 var
  l, i, l1, l2, dl: integer;
 begin
  l1:= -1;
  l2:= -1;
  for l:= 0 to length(lines) - 1 do with lines[l] do begin
   for i:= 0 to count - 1 do begin
    if (pos[i] = r1) then l1:= l;
    if (pos[i] = r2) then l2:= l;
   end;
   if (l1 > -1) and (l2 > -1) then break;
  end;
  l:= -1;
  if l1 > -1 then begin
   l:= l1;
   if (l2 > -1) and (l2 <> l1) then begin
    for i:= 0 to lines[l2].count - 1 do with lines[l1] do begin
     pos[count - 1]:= lines[l2].pos[i];
     inc(count);
    end;
    for dl:= l2 to length(lines) - 2 do begin
     lines[dl].count:= lines[dl + 1].count;
     for i:= 0 to lines[dl].count - 1 do
      lines[dl].pos[i]:= lines[dl + 1].pos[i];
    end;
    setLength(lines, length(lines) - 1);
    f:= False;
   end;
   if (l2 > -1) then exit;
  end else if l2 > -1 then l:= l2;
  if l < 0 then begin
   l:= length(lines);
   setLength(lines, l + 1);
   lines[l].count:= 0;
  end;
  if l1 < 0 then with lines[l] do begin
   pos[count]:= r1;
   inc(count);
  end;
  if l2 < 0 then with lines[l] do begin
   pos[count]:= r2;
   inc(count);
  end;
 end;

 begin
  setLength(lines, 0);
  Result:= False;
  repeat
   f:= True;
   for c:= 0 to fSCntr - 2 do if SPositions[c].Left > SPositions[c + 1].Left then begin
    f:= False;
    p:= SPositions[c];
    SPositions[c]:= SPositions[c + 1];
    SPositions[c + 1]:= p;
   end;
  until f;

  for c:= 0 to fSCntr - 2 do for n:= c + 1 to fSCntr - 1 do begin
   dx:= abs(SPositions[c].Left - SPositions[n].Left);
   dy:= abs(SPositions[c].Top - SPositions[n].Top);
   dh:= abs(SPositions[c].Height - SPositions[n].Height);
   ah:= (SPositions[c].Height + SPositions[n].Height) shr 1;
   if (dy < ah) and (dx < ah shl 1) and (dh < ah shr 1)
   then check(SPositions[c], SPositions[n]);
  end;
  lc:= 0;
  for n:= 0 to length(lines) - 1 do with lines[n] do begin
   if count in [36, 44] then begin
    if lc > 1 then begin
     setLength(lines, 0);
     exit(False);
    end;
    rLines[lc].minHeight:= 65535;
    rLines[lc].avgHeight:= 0;
    for c:= 0 to count - 1 do begin
     rLines[lc].pos[c]:= pos[c];
     if pos[c].Height < rLines[lc].minHeight
     then rLines[lc].minHeight:= pos[c].Height;
     inc(rLines[lc].avgHeight, pos[c].Height);
    end;
    rLines[lc].count:= count;
    rLines[lc].avgHeight:= round(rLines[lc].avgHeight / count);
    if rLines[lc].avgHeight shr 1 > rLines[lc].minHeight then begin
     setLength(lines, 0);
     exit(False);
    end;
    inc(lc);
   end;
  end;
  if (lc = 2) and (rLines[0].count = rLines[1].count) then begin
   fObjectsFound:= True;
   {$ifdef debugPassport}
   fSCntr:= 0;
   for lc:= 0 to 1 do with rLines[lc] do for c:= 0 to count - 1 do begin
    SPositions[fSCntr]:= pos[c];
    inc(fSCntr);
   end;
   {$endif}
   fText:= '';
   rTexts[0]:= '';
   rTexts[1]:= '';
   f:= rLines[0].pos[0].Top < rLines[1].pos[0].Top;
   with rLines[fli[f]] do avgHeight:= round(1.07 * avgHeight); // try compensate "<"
   angle:= - round(180 * arctan(
      (rectsK(rLines[0].pos[0], rLines[0].pos[rLines[0].count - 1]) +
       rectsK(rLines[0].pos[0], rLines[0].pos[rLines[0].count - 1])) / 2
     ) / pi);
   ah:= (rLines[0].avgHeight + rLines[1].avgHeight) shr 1;
   for lc:= 0 to 1 do with rLines[lc] do for c:= 0 to count - 1 do begin
    img1:= TgrayImage.Create;
    with pos[c] do img1.CopyRect(rect(left -2, Top - 2, Right + 2, Bottom + 2), img);
    if not img1.OneBit then img1.Convert1Bit;
    if fl[f, lc]
    then begin
     if c < 2 then s:= '<IP'
     else s:= '<ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    end else begin
     if c < 1
     then s:= '0123456789P'
     else if c < 9
     then s:= '<0123456789'
     else if c < 10
     then s:= '0123456789'
     else if c < 13
     then s:= '<ABCDEFGHIJKLMNOPQRSTUVWXYZ'
     else if c < 20
     then s:= '0123456789'
     else if c < 21
     then s:= 'FM'
     else if c < 10
     then s:= '0123456789'
     else if c < 28
     then s:= '0123456789'
     else begin
      if rLines[0].count = 36
      then begin
       if c < 35
       then s:= '<0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
       else s:= '<0123456789'
      end else begin
       if c < 42
       then s:= '<0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
       else s:= '<0123456789'
      end
     end
    end;
    rc:= SearchChar(img1, rect(2, 2, img1.Width - 2, img1.Height - 2), ah, angle, 100, s);
    img1.Free;
    if rc > #0 then rTexts[lc]:= rTexts[lc] + rc
    else exit(False);
   end;
   fText:= rTexts[fli[f]] + rTexts[fli[not f]];
   s:= rTexts[fli[f]];
   fDocType:= stringReplace(copy(s, 1, 2), '<', '', [rfReplaceAll]);
   Delete(s, 1, 2);
   fIssued:= stringReplace(copy(s, 1, 3), '<', '', [rfReplaceAll]);
   Delete(s, 1, 3);
   s:= stringReplace(s, 'KK', 'K', [rfReplaceAll]);
   s:= stringReplace(s, '<<', '<', [rfReplaceAll]);
   s:= stringReplace(s, '<K<', '', [rfReplaceAll]);
   if copy(s, length(s) - 2, 2) = '<K' then s:= copy(s, 1, length(s) - 2);
   s:= stringReplace(s, '<', ' ', [rfReplaceAll]);
   fSurname:= trim(s);
   s:= rTexts[fli[not f]];
   fNumber:= stringReplace(copy(s, 1, 9), '<', '', [rfReplaceAll]);;
   if not checksumIsOk(fNumber, s[10]) then exit(False);
   Delete(s, 1, 10);
   fNationality:= stringReplace(copy(s, 1, 3), '<', '', [rfReplaceAll]);
   Delete(s, 1, 3);
   t:= copy(s, 1, 6);
   if not checksumIsOk(t, s[7]) then exit(False);
   fBirthday:= StrToDate(copy(t, 1, 2) + '/' +copy(t, 3, 2) + '/' +copy(t, 5, 2), 'YY/MM/DD', '/');
   fSex:= s[8];
   Delete(s, 1, 8);
   t:= copy(s, 1, 6);
   if not checksumIsOk(t, s[7]) then exit(False);
   try
    fExpiration:= StrToDate(copy(t, 1, 2) + '/' +copy(t, 3, 2) + '/' +copy(t, 5, 2), 'YY/MM/DD', '/');
   except
    fExpiration:= date + 365 * 100 + 25;
   end;
   Delete(s, 1, 7);
   if rLines[0].count = 36
   then begin
    fPersonalNumber:= stringReplace(copy(s, 1, 7), '<', '', [rfReplaceAll]);
    t:= s[8];
    s:= rTexts[fli[not f]];
    s:= copy(s, 1, 10) + copy(s, 14, 7) + copy(s, 22, 7);
    if not checksumSIsOk(checksumMod(s) + checksumMod(fPersonalNumber), t[1]) then exit(False);
   end else begin
    fPersonalNumber:= stringReplace(copy(s, 1, 14), '<', '', [rfReplaceAll]);
    if not checksumIsOk(fPersonalNumber, s[15]) then exit(False);
    s:= rTexts[fli[not f]];
    t:= s[44];
    s:= copy(s, 1, 10) + copy(s, 14, 7) + copy(s, 22, 22);
    if not checksumIsOk(s, t[1]) then exit(False);
   end;
   Result:= True;
  end;
 end;

begin
 inherited;
 fObjectsFound:= False;
 fFound:= False;
 Result:= False;
 fSCntr:= 0;
 fSHeightN:= 255;
 fSHeightX:= 0;
 for c:= 0 to length(Images) - 1 do begin
  if Images[c].Height > fSHeightX then fSHeightX:= Images[c].Height;
  if Images[c].Height < fSHeightN then fSHeightN:= Images[c].Height;
 end;
 if fSHeightX = 0 then exit(False);
 fSHeightN:= round(fSHeightN * 0.8); // "<" symbol height
 fSHeightX:= round(fSHeightX * 1.1) + 1;
 df.DateSeparator:= #0;
 df.ShortDateFormat:= 'YYMMDD';
 df.LongDateFormat:= df.ShortDateFormat;

 if badVideo then img.Normalize1Bit(fSHeightX * 3 shr 1);

 img.FindObjects(@CheckSymbols, r);

 if fSCntr > 70 then begin
  Result:= tryRecognize;
  if not Result and not img.OneBit then begin
   fc:= 0;
   for c:= 0 to length(lines) - 1 do
    if lines[c].count > 20 then inc(fc);
   if fc > 1 then begin
    img.Normalize1Bit(fSHeightX * 3 shr 1);
    fSCntr:= 0;
    img.FindObjects(@CheckSymbols, r);
    if fSCntr > 70 then Result:= tryRecognize;
   end;
  end;
 end;
 setLength(lines, 0);
 fFound:= Result;
end;

procedure TpassportFinder.CheckSymbols(const Sender: TgrayImage; const r: TRect; var stop: boolean);
var
 w, h: word;
 x, y, c: integer;
begin
 w:= r.Right - r.Left;
 h:= r.Bottom - r.Top;
 if (fSHeightX > h) and (fSHeightN < h) and (fSHeightX > w) then begin
  c:= 0;
  for x:= r.Left to r.Right do for y:= r.Top to r.Bottom do
   if Sender.Pixel[x, y] < 128 then inc(c);
  if c < (w + h) then exit;
  for c:= 0 to fSCntr - 1 do with SPositions[c] do
   if (Left <= r.Right) and (r.Left <= Right) and (Top <= r.Bottom) and (r.Top <= Bottom)
   then exit;
  if h > w then begin
   SPositions[fSCntr]:= r;
   inc(fSCntr);
  end;
 end;
end;

end.

