unit vvvRegNum;

{$mode objfpc}{$H+}

interface

uses
 Classes, SysUtils, vvvOCR;

type
 TplateFinder = class(TbasicOCR)
 private
  fSHeightN, fSHeightX: byte;
  fSCntr: byte;
  fFound: boolean;
  fInverse: boolean;
  fObjectsFound: boolean;
  procedure CheckPlate(const Sender: TgrayImage; const r: TRect; var stop: boolean);
  procedure CheckSymbols(const Sender: TgrayImage; const r: TRect; var stop: boolean);
 public
  {$ifdef debug}
  minW, maxW, avgW, rAngle: integer;
  rRate: byte;
  SPOk: array [byte] of Boolean;
  {$endif}
  minWidth, minHeight, maxWidth, maxHeight: word;
  minKWH, maxKWH: byte;
  SPositions: array [byte] of TRect;
  PlatePosition: TRect;
  tryInversion: boolean;
  tryMaxResult: boolean;
  tryRatio: boolean;
  tryAngle: boolean;
  property SCntr: byte read fSCntr;
  property Found: boolean read fFound;
  property ObjectsFound: boolean read fObjectsFound;
  property Inversion: boolean read fInverse;
 public
  constructor Create;
  function Search(const img: TgrayImage; const r: TRect): boolean; override;
 end;

implementation

constructor TplateFinder.Create;
begin
 inherited;
 minWidth:= 80;
 minHeight:= 16;
 maxWidth:= 200;
 maxHeight:= 40;
 minKWH:= 35;
 maxKWH:= 55;
 minWeight:= 75;
 tryInversion:= False;
 tryMaxResult:= False;
 tryRatio:= True;
 tryAngle:= False;
end;

function TplateFinder.Search(const img: TgrayImage; const r: TRect): boolean;
begin
 inherited;
 fObjectsFound:= False;
 fFound:= False;
 fInverse:= False;
 img.FindObjects(@CheckPlate, r, not fInverse);
 if tryInversion and not fObjectsFound then begin
  fInverse:= True;
  img.FindObjects(@CheckPlate, r, not fInverse);
 end;
 Result:= fFound;
end;

procedure TplateFinder.CheckPlate(const Sender: TgrayImage; const r: TRect; var stop: boolean);
const
 aAlpha: ansiString = 'ABCDEFGHIJKLMN0PRSTUVXYZ';
 aNumeric: ansiString = '0123456789';
 aRatio: array [0..4] of byte = (100, 95, 105, 90, 110);
 aAngles: array [0..8] of integer = (0, 1, -1, 2, -2, 3, -3, 4, -4);
 rFrame = 2;

var
 w, h, h1, h2: word;
 k: cardinal;
 i, c, s, rt: byte;
 storedMinWeight: byte;
 f: boolean;
 p: TRect;
 aH: cardinal;
 xImg: TgrayImage;

 function xNormal(x: integer): ansiString;
 begin
  if x > 3 then Result:= aNumeric
  else if SPositions[s].Left + h2 < SPositions[s + 1].Left then begin
   if x = 0
   then Result:= 'HMP' + aNumeric
   else if (x in [2,3]) and (length(fText) > 0) and (fText[1] < 'A')
   then Result:= aAlpha
   else Result:= aNumeric;
  end else if x in [2,3] then Result:= aAlpha else Result:= aNumeric;
 end;

 function xInverse(x: integer): ansiString;
 begin
  if (SPositions[1].Left + h2 < SPositions[2].Left) and (SPositions[3].Left + h2 < SPositions[4].Left)
  then begin // blue
   if x = 2 then Result:= 'T'
   else Result:= xNormal(x)
  end else if x > 2 then Result:= aNumeric else begin
   if SPositions[0].Left + h2 < SPositions[1].Left then begin // space after first char
    if x = 0 then Result:= 'DT' else Result:= aNumeric;
   end else if SPositions[1].Left + h2 < SPositions[2].Left then begin // FK-
    case x of
    0: Result:= 'F';
    1: Result:= 'K';
    else Result:= aNumeric;
    end;
   end else case x of // SFR
    0: Result:= 'S';
    1: Result:= 'F';
    2: Result:= 'R';
    end;
  end;
 end;

 function xCorrect(x: integer; ch: ansiChar; var sOk: boolean): ansiString;
 begin
  if (x in [2,3]) and (length(fText) > 0) and (fText[1] < 'A') and (ch = '0') then Result:= 'O'
  else if ch = #0 then Result:= '' else Result:= ch;
  sOk:= (Result > '');
  {$ifdef debug}
  if minW > CharWeight then minW:= CharWeight;
  if maxW < CharWeight then maxW:= CharWeight;
  inc(avgW, CharWeight);
  SPOk[x]:= SPOk[x] or sOk;
  {$endif}
 end;

 procedure ClearVars;
 begin
  fText:= '';
  {$ifdef debug}
  minW:= 65535;
  maxW:= 0;
  avgW:= 0;
  rRate:= 100;
  rAngle:= 0;
  {$endif}
 end;

begin
 w:= r.Right - r.Left;
 h:= r.Bottom - r.Top;
 k:= w * 10 div h;
 ClearVars;
 if (h >= minHeight) and (h <= maxHeight) and (w >= minWidth) and (w <= maxWidth) and (k >= minKWH) and (k <= maxKWH)
 then begin
  {$ifdef debug}
  for i:= 0 to 255 do SPOk[i]:= False;
  {$endif}
  fSHeightX:= h - h shr 3; // 7/8
  fSHeightN:= h shr 1; // 1/2
  fSCntr:= 0;
  xImg:= TgrayImage.Create;
  xImg.CopyRect(rect(r.Left - rFrame, r.Top - rFrame, r.Right + rFrame, r.Bottom + rFrame), Sender);
  //xImg.Normalize;
  xImg.FindObjects(@CheckSymbols, rect(0, 0, xImg.Width, xImg.Height), fInverse);
  if (fSCntr in [7..8]) and not fInverse or fInverse and (fSCntr in [6..7]) then begin
   // sort symbols
   repeat
    f:= True;
    for c:= 0 to fSCntr - 2 do
     if SPositions[c].Left > SPositions[c + 1].Left then begin
      f:= False;
      p:= SPositions[c + 1];
      SPositions[c + 1]:= SPositions[c];
      SPositions[c]:= p;
     end;
   until f;

   if fSCntr = 8 then s:= 1 else s:= 0;
   aH:= 0;
   h1:= 65535;
   h2:= 0;
   for c:= s to fSCntr - 1 do with SPositions[c] do begin
    inc(aH, Height);
    if h1 > Height then h1:= Height;
    if h2 < Height then h2:= Height;
   end;
   aH:= aH div (fSCntr - s);

   if h2 - h1 - 1 > aH shr 3 then begin
    xImg.Free;
    exit;
   end;

   h2:= aH * 6 shr 3;

   fObjectsFound:= True;

   xImg.Convert1Bit(fInverse);

   storedMinWeight:= minWeight - 5;
   if tryMaxResult then minWeight:= 90;
   repeat
    i:= 0;
    repeat
     rt:= 0;
     repeat
      ClearVars;
      if fInverse
      then for c:= s to fSCntr - 1 do begin
       fText:= fText +
        xCorrect(c, SearchChar(xImg, SPositions[c], aH, aAngles[i], aRatio[rt], xInverse(c)), f);
        if tryMaxResult and not f then break;
      end else for c:= s to fSCntr - 1 do begin
       fText:= fText +
        xCorrect(c-s, SearchChar(xImg, SPositions[c], aH, aAngles[i], aRatio[rt], xNormal(c - s)), f);
       if tryMaxResult and not f then break;
      end;

      fFound:= length(fText) = (fSCntr - s);

      {$ifdef debug}
      if fFound then begin
       rRate:= aRatio[rt];
       rAngle:= aAngles[i];
      end;
      {$endif}

      inc(rt);
     until fFound or (rt > 4) or not tryRatio;
     inc(i);
    until fFound or (i > 8) or not tryAngle;
    dec(minWeight, 5);
   until fFound or (minWeight < storedMinWeight);

   minWeight:= storedMinWeight + 5;

   {$ifdef debug}
   c:= 0;
   for i:= 0 to 255 do if SPOk[i] then inc(c);
   if c > 0 then avgW:= avgW div c else begin
     avgW:= 0;
     maxW:= 0;
   end;
   if avgW = 0 then minW:= 0;
   {$endif}

   for c:= 0 to fSCntr - 1 do with SPositions[c] do begin
    inc(Left, r.Left - rFrame);
    inc(Right, r.Left - rFrame);
    inc(Top, r.Top - rFrame);
    inc(Bottom, r.Top - rFrame);
   end;
   PlatePosition:= r;
  end;
  xImg.Free;
 end;
 stop:= fFound;
end;

procedure TplateFinder.CheckSymbols(const Sender: TgrayImage; const r: TRect; var stop: boolean);
var
 w, h: word;
 c: integer;
begin
 w:= r.Right - r.Left;
 h:= r.Bottom - r.Top;
 if (fSHeightX > h) and (fSHeightN < h) and (h > w) then begin
  for c:= 0 to fSCntr - 1 do with SPositions[c] do
   if (Left <= r.Left) and (Right > r.Left) and (Top <= r.Top) and (Bottom > r.Top)
   then exit;
  SPositions[fSCntr]:= r;
  inc(fSCntr);
 end;
end;

end.

