unit vvvTrainingOCR;

{$mode objfpc}{$H+}

interface

uses
 Classes, SysUtils, vvvOCR;

type
 TTrainingOCR = class(TbasicOCR)
 private
  oGraySum: TgrayImage;
  function oSearchChars(img: TgrayImage; r: TRect; h: byte; a: integer;
    p: byte; valid: ansiString): ansiString;
 public
  oSimilarChars: ansiString;
  oSize: integer;
  oAngle: integer;
  oRatio: integer;
  oSymbol: ansiChar;
  oGraySymbol: TgrayImage;
 public
  procedure AddSymbol(aHeight: word; aAngle: integer; aRatio: byte;
    aSymbol: char; aImg: TgrayImageData); overload;
  procedure AddSymbol(aHeight: word; aAngle: integer; aRatio: byte;
    aSymbol: char; aImg: TgrayImage); overload;
  procedure SaveToStream(s: TStream);
  procedure Optimize(cbp: TNotifyEvent; gc: TGrayImageClass = nil);
 end;


function abDataAsGrayImage(dt: TgrayImageData; gc: TGrayImageClass; add: byte = 0): TgrayImage;

implementation

type
 TinternalGrayImage = class(TgrayImage)
 public
  property rawData;
  property Data;
 end;

const
 ocrColors: array [0..4] of byte = (0, 64, 128, 192, 255);

procedure TTrainingOCR.AddSymbol(aHeight: word; aAngle: integer; aRatio: byte; aSymbol: char; aImg: TgrayImageData); overload;
var
 c: integer;
 img: PabImageData;
 imgs: TabImages;
begin
 imgs:= nil;
 for c:= 0 to length(Images) - 1 do with Images[c] do begin
  if (Height = aHeight) and (Angle = aAngle) and (Ratio = aRatio) then imgs:= Self.Images[c];
 end;
 if not assigned(imgs) then begin
  setLength(Images, length(Images) + 1);
  Images[length(Images) - 1]:= TabImages.Create(aHeight, aAngle, aRatio);
  imgs:= Images[length(Images) - 1];
 end;
 img:= nil;
 for c:= 0 to imgs.Count - 1 do with imgs.Images[c] do
  if (symbol = aSymbol) and (image.Width = aImg.Width)
  then img:= @imgs.Images[c];
 if not assigned(img) then begin
  setLength(imgs.Images, length(imgs.Images) + 1);
  img:= @imgs.Images[length(imgs.Images) - 1];
 end;
 img^.symbol:= aSymbol;
 img^.maxWeight:= 0;
 img^.image.Width:= aImg.Width;
 img^.image.Height:= aImg.Height;
 setLength(img^.image.Data, length(aImg.Data));
 move(aImg.Data[0], img^.image.Data[0], length(aImg.Data));
 with img^, img^.image do for c:= 0 to length(Data) - 1 do begin
  if Data[c] <= 32 then begin
   inc(maxWeight, ocrWeights[0, 0]);
   Data[c]:= 0;
  end else if Data[c] <= 64 then begin
   inc(maxWeight, ocrWeights[0, 1]);
   Data[c]:= 1;
  end else if Data[c] <= 128 then begin
   inc(maxWeight, ocrWeights[0, 2]);
   Data[c]:= 2;
  end else if Data[c] <= 192 then begin
   inc(maxWeight, ocrWeights[1, 3]);
   Data[c]:= 3;
  end else begin
   inc(maxWeight, ocrWeights[1, 4]);
   Data[c]:= 4;
  end;
  aImg.Data[c]:= ocrColors[Data[c]];
 end;
end;

procedure TTrainingOCR.AddSymbol(aHeight: word; aAngle: integer; aRatio: byte; aSymbol: char; aImg: TgrayImage); overload;
begin
 AddSymbol(aHeight, aAngle, aRatio, aSymbol, TinternalGrayImage(aImg).rawData);
end;

procedure TTrainingOCR.SaveToStream(s: TStream);
var
 i: integer;
 c: word;
begin
 c:= length(Images);
 s.Write(c, SizeOf(c));
 for i:= 0 to c - 1 do Images[i].SaveToStream(s);
end;

function TTrainingOCR.oSearchChars(img: TgrayImage; r: TRect; h: byte; a: integer;
  p: byte; valid: ansiString): ansiString;
var
 ci, cs, x, y, sx, sy, ax, ay, yw,
 tw, xw, ssx, ssy: integer;

 function getPixel(Data: TgrayImageData; x, y: integer): byte;
 begin
  if (x >= Data.Width) or (y >= Data.Height) or (x < 0) or (y < 0)
  then Result:= 4
  else Result:= Data.Data[Data.Width * y + x];
 end;
begin
 Result:= '';
 if r.Top > 2 then ay:= r.Top - 3 else ay:= 0;
 if r.Left > 2 then ax:= r.Left - 3 else ax:= 0;
 for ci:= 0 to length(Images) - 1 do with Images[ci] do begin
  if (Height = h) and (Angle = a) and (Ratio = p) then begin
   for cs:= 0 to length(Images) - 1 do with Images[cs] do begin
    if (abs(image.Width - 2 - r.Width) < 3) and (pos(symbol, valid) > 0)
    then begin
     xw:= 0;
     for sx:= ax to r.Left + 1 do for sy:= ay to r.Top + 1 do begin
      tw:= 0;
      for y:= 0 to image.Height - 1 do begin
       yw:= y * image.Width;
       for x:= 0 to image.Width - 1
        do inc(tw, ocrWeights[img.Pixel[sx + x, sy + y], image.Data[x + yw]]);
      end;
      if tw > xw
      then begin
       xw:= tw;
       ssx:= sx - 2;
       ssy:= sy - 2;
      end;
     end;
     xw:= xw * 100 div maxWeight;
     if xw > minWeight then begin
      Result:= Result + symbol;
      for sy:= 0 to oGraySum.Height - 1 do for sx:= 0 to oGraySum.Width - 1 do begin
       oGraySum.Pixel[sx, sy]:= oGraySum.Pixel[sx, sy] + getPixel(image, sx + ssx, sy + ssy);
      end;
     end;
    end;
   end;
  end;
 end;
 xw:= length(Result);
 if xw > 1 then begin
  for sy:= 0 to oGraySum.Height - 1 do for sx:= 0 to oGraySum.Width - 1 do begin
   oGraySum.Pixel[sx, sy]:= oGraySum.Pixel[sx, sy] div xw;
  end;
 end;
end;

procedure TTrainingOCR.Optimize(cbp: TNotifyEvent; gc: TGrayImageClass = nil);
const
 add = 2;
var
 ms: TMemoryStream;
 tmpOCR: TTrainingOCR;
 abCntr: integer;
 c, sc: integer;
 validChars: ansiString;
 f: ansiChar;
begin
 tmpOCR:= TTrainingOCR.Create;
 ms:= TMemoryStream.Create;
 try
  for abCntr:= 0 to length(Images) - 1 do with Images[abCntr] do begin
   oAngle:= Angle;
   oRatio:= Ratio;
   oSize:= Height;
   for sc:= 0 to length(Images) - 1 do begin
    validChars:= '';
    oSimilarChars:= '';
    oSymbol:= Images[sc].symbol;
    for c:= 0 to length(Images) - 1 do if c <> sc
     then validChars:= validChars + Images[c].symbol;
    oGraySymbol:= abDataAsGrayImage(Images[sc].image, gc, add);
    oGraySum:= TgrayImage.Create;
    oGraySum.Width:= Images[sc].image.Width;
    oGraySum.Height:= Images[sc].image.Height;
    oGraySymbol.Convert1Bit;
    oSimilarChars:= oSearchChars(oGraySymbol,
      rect(add, add,
        add + Images[sc].image.Width,
        add + Images[sc].image.Height),
      oSize, oAngle, oRatio, validChars);
    try
     if length(oSimilarChars) > 0 then begin
      with TinternalGrayImage(oGraySymbol) do begin
       Copy(Images[sc].image);
       for c:= 0 to length(Data) - 1 do begin
        if (Data[c] = 2) and (TinternalGrayImage(oGraySum).Data[c] > 2)
        then Data[c]:= 1
        else if (Data[c] = 3) and (TinternalGrayImage(oGraySum).Data[c] < 2)
        then Data[c]:= 4;
        Data[c]:= ocrColors[Data[c]];
       end;
      end;
      tmpOCR.AddSymbol(oSize, oAngle, oRatio, oSymbol, oGraySymbol);
      if assigned(cbp) then cbp(self);
     end else with TinternalGrayImage(oGraySymbol) do begin
      Copy(Images[sc].image);
      for c:= 0 to length(Data) - 1 do Data[c]:= ocrColors[Data[c]];
      tmpOCR.AddSymbol(oSize, oAngle, oRatio, oSymbol, oGraySymbol);
     end;
    finally
     oGraySymbol.Free;
     oGraySum.Free;
    end;
   end;
  end;
  tmpOCR.SaveToStream(ms);
  ms.Position:= 0;
  ClearSymbols;
  LoadFromStream(ms);
 finally
  tmpOCR.Free;
  ms.Free;
 end;

end;

function abDataAsGrayImage(dt: TgrayImageData; gc: TGrayImageClass; add: byte = 0): TgrayImage;
var
 x, y, c: integer;
begin
 if not assigned(gc) then gc:= TgrayImage;
 Result:= gc.Create;
 Result.Width:= dt.Width + add * 2;
 Result.Height:= dt.Height + add * 2;
 with TinternalGrayImage(Result) do begin
  for c:= 0 to length(Data) - 1 do Data[c]:= 4;
  c:= 0;
  for y:= add to Height - add - 1 do for x:= add to Width - add - 1 do begin
   Pixel[x, y]:= dt.Data[c];
   inc(c);
  end;
  for c:= 0 to length(Data) - 1 do Data[c]:= ocrColors[Data[c]];
 end;
end;

end.

